#article title:  Effects of stopping criterion on the growth of trees in regression random forests
#journal name: The New England Journal of Statistics in Data Science (NEJSDS)
#author names: Arsham, Rosenberg, Little
#affiliation of the corresponding author: Goucher College
#e-mail of the corresponding author: aryana.arsham@goucher.edu


################################################################### Function for regression RF

#### The following function provides predictions for testing set and corresponding variable importances.

####### Inputs: M, m_try, node_size, training_data, testing_data, approach, seed_num

# Format of data: Data should be provided in a data.frame having the following structure.
# The first column are observation weights. If weights are not included in the data, then this column should be entries all equal to one.
# Weights for an observation need to represent the number of units in the population represented by that observation, such that the sum of all weights is the population size.
# The second column corresponds to the outcome (dependent variable), this variable needs to be continuous
# The remaining columns corresponds to input variables (independent variables) used to predict the outcome
# Testing and training data need to be in two separate files
# Testing and training set data also need to have common names and order for all input variables

# M: number of trees in RF model
# m_try: Number of features randomly sampled (without replacement) at each node split
# node_size: minimum number of observations in (parent nodes for parent approach) leaf nodes for leaf approach
# training_data: data set that will be sampled to create RF model, this is specified by used prior to using functions (see example illustration)
# testing_data: data set that will be used to assess the performance of the RF algorithm. Outcomes from this dataset are predicted
# approach: select either "leaf" or "parent" approach
# seed_num: value used for setting the seed

####### Outputs:

# Estimated generalization error: Weighted MSE from testing set.
# Estimated variable importances: input variables and associated variable importance values from RF model
# Predicted testing outcome: vector of predicted testing outcomes
# Testing outcome: observed testing outcomes
# m_try: m_try hyperparameter specified by user
# node size: node size hyperparameter specified by user
# M: number of regression trees in forest, specified by user
# Approach used: RF approach used  ("leaf" or "parent"), specified by user
# Variable importance plot

######### Additional plots for model checking can found with the illustration

############# Important note: The outcomes from training and testing sets should have any missing values imputed prior to applying the function.


runners <- function(M, m_try, node_size, training_data, testing_data, approach, seed_num )

    #M = number of trees
    #m_try = number of sampled features at each node for splitting
    #node_size = minimum node size
    #training_data = training data
    #testing_data = test data
    #approach = "leaf" or "parent" depending on whether leaf or parent node is used to limit terminating node size
    #seed_num = random number seed - for reproducibility of results

{
    #Load R packages
    rm(list=ls())
    library(plyr)
    library(stringr)
    library(zoo)
    library(riskRegression)
    library(ggplot2)
    options(warn=-1)
    
    # Splitting algorithm for leaf approach
    sse_var_leaf = function(x,y, ww, node_size)
    {
        splits=  rollapply(sort(unique(x)), 2, mean, by = 1, align = "left", partial = F)
        sse = c()
        splits_pot = c()
        splits_pot_min = c()
        
        for (i in seq_along(splits)) {
            sp = splits[i]
            sse[i] = sum(ww[x < sp] *(y[x < sp] - mean(y[x < sp]))^2) + sum(ww[x >= sp] *(y[x >= sp] - mean(y[x >= sp]))^2)
            splits_pot[i] = min(length(y[x >= sp]), length(y[x < sp]))
            
        }
        splits_pot_min = (splits_pot >= node_size)
        splits_pot = splits[splits_pot_min == TRUE]
        splits_pot_sse = sse[splits_pot_min == TRUE]
        split_at = splits_pot[which.min(splits_pot_sse)]
        #return minimal MSE and minimum node size at that point
        return(list(sse = min(splits_pot_sse), split = split_at))
    }
    
    # Splitting algorithm for parent approach
    sse_var_parent = function(x,y, ww, node_size)
    {
        splits=  rollapply(sort(unique(x)), 2, mean, by = 1, align = "left", partial = F)
        sse = c()
        splits_pot = c()
        splits_pot_min = c()
        
        for (i in seq_along(splits)) {
            sp = splits[i]
            sse[i] = sum(ww[x < sp] *(y[x < sp] - mean(y[x < sp]))^2) + sum(ww[x >= sp] *(y[x >= sp] - mean(y[x >= sp]))^2)
            splits_pot[i] = length(y[x >= sp])+length(y[x < sp])
            
        }
        splits_pot_min = (splits_pot >= node_size)
        splits_pot = splits[splits_pot_min == TRUE]
        splits_pot_sse = sse[splits_pot_min == TRUE]
        split_at = splits_pot[which.min(splits_pot_sse)]
        #return minimal MSE and minimum node size at that point
        return(list(sse = min(splits_pot_sse), split = split_at))
    }
    
    
    
    
    
    # The reg_tree function creates a regression tree
    
    reg_tree <- function(formula_train , data , data_train, node_size , features, m_try, outcome, data_test, approach) {
        
        #formula_train = formula for model fit
        #data = bootstrap sampled training data (sampled from data_train)
        #data_train = training data
        #data_test = test data
        #node_size = minimum node size
        #features = names of features (independent variables) in data
        #m_try = number of sampled features at each node for splitting
        #outcome = name of outcome (dependent) variable
        #approach = "leaf" or "parent" depending on whether leaf or parent node is used to limit terminating node size (=node_size)
        
        
        # Design matrix for training set
        X <- model.matrix(formula_train, data)
        
        # Outcome for training set
        y<- data[,2]
        
        # Weights for training set
        wei <-data[,1]
        
        # Design matrix for testing set
        X.test <- model.matrix(formula_train, data_test)
        
        # Outcome for testing set
        y.test <- data_test[,2]
        
        # Initialize condition for node splitting as TRUE
        continue_splits <- TRUE
        
        # Create output data.frame with splitting rules and observations for training set
        tree_info <- data.frame(NODE = 1, NOBS = nrow(data), FILTER = NA, TERMINAL = "SPLIT",
        F_IMP = NA, SPLIT = NA, stringsAsFactors = FALSE)
        
        # Create output data.frame with splitting rules and observations for testng set
        tree_info_test <- data.frame(NODE = 1, NOBS = nrow(data_test), FILTER = NA,TERMINAL = "SPLIT",stringsAsFactors = FALSE)
        
        # Keep splitting until there are only leaf nodes
        while(continue_splits) {
            
            # Determine parents that need to be split
            to_calculate <- which(tree_info$TERMINAL == "SPLIT")
            
            # Number of parent nodes to be split
            length_to_calculate = length(to_calculate)
            
            # If no parent nodes need to be split exit
            if(length_to_calculate ==0)
            {
                break
            }
            
            # For loop for splitting nodes
            for (j in to_calculate) {
                
                
                # Partitioning training set
                if (!is.na(tree_info[j, "FILTER"]))
                {
                    # Subset data according to the current filter
                    this_data <- subset(data, eval(parse(text = tree_info[j, "FILTER"])))
                    # Get the design matrix
                    X <- model.matrix(formula_train, this_data)
                }else{
                    # For root node usetraining set (no filter yet)
                    this_data <- data
                }
                
                
                
                # Partitioning testing set
                if (!is.na(tree_info_test[j, "FILTER"]))
                {
                    # Subset data according to the current filter
                    this_data_test <- subset(data_test, eval(parse(text = tree_info_test[j, "FILTER"])))
                    # Get the design matrix
                    X.test <- model.matrix(formula_train, this_data_test)
                }else{
                    # For root node usetraining set (no filter yet)
                    this_data_test <- data_test
                }
                
                
                
                # Sample m_try number of input variables without replacement
                features_sample <- sample(features, size = m_try, replace = F)
                
                # Create new formula for training the RF model that only includes the m_try input variables sampled from previous line of code
                formula_new <- as.formula(paste0(outcome, " ~ 0 +  ", paste0(features_sample, collapse =  " + ")))
                
                
                # Design matrix which includes sampled input variables
                X.split = model.matrix(formula_new, this_data)
                
                # Apply node splitting depending on approach specified by user
                if( approach == "leaf")
                {
                    splitting <- apply(X.split,  MARGIN = 2, FUN = sse_var_leaf, y = this_data[,2],ww=this_data[,1], node_size=node_size)
                }else{
                    splitting <- apply(X.split,  MARGIN = 2, FUN = sse_var_parent, y = this_data[,2],ww=this_data[,1], node_size=node_size)
                }
                
                # Store input variable and their associated optimal cut-point
                splittings= as.matrix(splitting)
                splitting= do.call("cbind", splittings[,1])
                
                # Detect if any splits were selected
                length_inf= length(which(splitting[1,] == Inf))
                
                # If no splits were selected set current nodes as leaf nodes (terminal nodes of tree)
                if(length_inf == m_try)
                {
                    tree_info[j, "TERMINAL"] <- "LEAF"
                    tree_info_test[j, "TERMINAL"] <- "LEAF"
                    split_here <- rep(FALSE, 2)
                    split_here_test <- rep(FALSE, 2)
                    next
                }
                
                
                
                
                # Select variable and associated cut-point having largest reduction in SSE
                tmp_splitter <- which.min(splitting[1,])
                
                
                # Count number of nodes in tree
                max_node <- max(tree_info$NODE)
                
                # Get current filter based on optimal variable and cut-point just determined
                current_filter <- c(paste( noquote(colnames( X.split)[ tmp_splitter ]), ">=",
                splitting[2,tmp_splitter]),
                paste(noquote(colnames( X.split)[ tmp_splitter ]), "<",
                splitting[2,tmp_splitter]))
                
                # Use current filter for testing set
                current_filter_test = current_filter
                
                # Check current filter has been applied for training set
                split_here  <- !sapply(current_filter,
                FUN = function(x,y) any(grepl(x, x = y)),
                y = tree_info$FILTER)
                
                # Check current filter has been applied for testing set
                split_here_test  <- !sapply(current_filter_test,
                FUN = function(x,y) any(grepl(x, x = y)),
                y = tree_info_test$FILTER)
                
                # Attach current filter to training set tree
                if (!is.na(tree_info[j, "FILTER"])) {
                    current_filter  <- paste(tree_info[j, "FILTER"],
                    current_filter, sep = " & ")
                }
                
                
                # Attach current filter to testing set tree
                if (!is.na(tree_info_test[j, "FILTER"])) {
                    current_filter_test  <- paste(tree_info_test[j, "FILTER"],
                    current_filter_test, sep = " & ")
                }
                
                #Function for obtaining weighted SSE
                sse_imp <- lapply(current_filter,
                FUN = function(i, x)
                {
                    df <- subset(x = x, subset = eval(parse(text = i)))
                    cy <- df[,2]
                    cwei <- df[,1]
                    imp <- sum(cwei*(cy - mean(cy, na.rm = TRUE))^2)
                    return(c(imp))
                },
                x = this_data)
                
                
                # Get the number of observations in newly created children nodes for training data
                current_nobs = sapply(current_filter,FUN = function(i, x) {nrow(subset(x = x, subset = eval(parse(text = i))))},x = this_data)
                
                # Determine weighted SSE from each newly created children nodes, then take sum
                imp_sum_child <- sum(sapply(sse_imp, function(x) x[[1]]))
                
                # Parent node outcomes
                current_y <- this_data[,2]
                
                # Weights in parent node
                wei_y <- this_data[,1]
                
                # Determine SSE in parent node
                imp_parent <- sum(wei_y*(current_y-mean(current_y))^2)
                
                # Reduction in SSE from new split
                f_imp <- imp_parent - imp_sum_child
                
                
                # Get the number of observations in newly created children nodes for testing data
                current_nobs_test = sapply(current_filter_test,FUN = function(i, x) {nrow(subset(x= x, subset = eval(parse(text = i))))},x= this_data_test)
                
                # If any of the new children nodes in training set have fewer than node size observations and leaf approach is selected
                if (any(current_nobs < node_size)  & (approach == "leaf"))
                {
                    # If any children nodes have fewer than node size observations register that they should not be split
                    min_no= which(current_nobs < node_size)
                    split_here[min_no] <- FALSE
                    split_here_test[min_no] <- FALSE
                }
                
                # If the sum of the observations in the new children nodes is fewer than node size observations and parent approach is selected
                 if ( (sum(current_nobs) < node_size)& (approach == "parent"))
                {
                    #If any children nodes have fewer than node size observations register that they should not be split
                    split_here[1:2] = FALSE
                    split_here_test[1:2] = FALSE
                }
                

                # Update newly created children nodes for training set
                children <- data.frame(NODE = c(max_node+1, max_node+2),
                NOBS = current_nobs,
                FILTER = current_filter,
                TERMINAL = rep("SPLIT", 2),
                F_IMP = NA,
                SPLIT = NA,
                row.names = NULL)[split_here,]
                
                # Update newly created children nodes for testing set
                children_test <- data.frame(NODE = c(max_node+1, max_node+2),
                NOBS = current_nobs_test,
                FILTER = current_filter_test,
                TERMINAL = rep("SPLIT", 2),
                SPLIT = NA,
                row.names = NULL)[split_here_test,]
                
                # Update tree for training set
                tree_info[j, "TERMINAL"] <- ifelse(all(!split_here), "LEAF", "PARENT")
                tree_info[j, "F_IMP"] <- f_imp
                if (tree_info[j, "TERMINAL"] == "PARENT")
                {
                    tree_info[j, "SPLIT"] <- noquote(colnames( X.split)[ tmp_splitter ])
                }
                
                # Update tree for testing set
                tree_info_test[j, "TERMINAL"] <- ifelse(all(!split_here_test), "LEAF", "PARENT")
                if (tree_info_test[j, "TERMINAL"] == "PARENT")
                {
                    tree_info_test[j, "SPLIT"] <- noquote(colnames( X.split)[ tmp_splitter ])
                }
                
                
                # Add newly created children nodes to tree for training set
                tree_info <- rbind(tree_info, children)
                
                
                # Add newly created children nodes to tree for testing set
                tree_info_test <- rbind(tree_info_test, children_test)
                
                # Check if there are any splits left in tree
                continue_splits <- !all(tree_info$TERMINAL != "SPLIT")
                
            } # end for
            
        } # end while
        
        
        # Determine which nodes are leaves in regression tree for the training set
        leafs <- tree_info[tree_info$TERMINAL == "LEAF", ]
        
        # Determine which nodes are leaves in regression tree for the testing set
        leafs_test <- tree_info_test[tree_info_test$TERMINAL == "LEAF", ]
        
        # Create empty array for storing predicted values from training set
        fitted <- c()
        
        # Create empty array for storing predicted values from testing set
        fitted_test<- c()
        
        # For loop for each leaf node
        for (i in seq_len(nrow(leafs)))
        {
            
            # Index of observations in current leaf node for training set
            ind <- as.numeric(rownames(subset(data, eval(parse(text = leafs[i, "FILTER"])))))
            ind_f = ind[1]
            
            # Index of observations in current leaf node for testing set
            ind_test = as.numeric(rownames(subset(data_test, eval(parse(text = leafs_test[i, "FILTER"])))))
            
            # For training set: Assign all observations in current leaf node with prediction equal to weighted average of all training outcomes in current leaf node
            fitted[ind] <- sum((data[,1])[ind]* (data[,2])[ind])/sum((data[,1])[ind])
            
            # For testing set: Assign all observations in current leaf node with prediction equal to weighted average of all training outcomes in current leaf node
            fitted_test[ind_test] <-  fitted[ind_f]
        }
        # Above for loop computed the predictions for each observation based on the regression tree model created
        
        
        # Extract variables and corresponding reduction in SSE by splitting
        imp <- tree_info[, c("SPLIT", "F_IMP")]
        
        # For each variable take sum of all of its associated reduction in SSE
        if (!all(is.na(imp$SPLIT)))
        {
            imp <- aggregate(F_IMP ~ SPLIT, FUN = function(x, all) sum(x, na.rm = T),
            data = imp, all = imp$F_IMP)
        }
        
        
        # Denote input variables by "FEATURES" and associated importances from built tree model as "IMPORTANCE"
        names(imp) <- c("FEATURES", "IMPORTANCE")
        
        # Sort importances in decreasing order
        imp <- imp[order(imp$IMPORTANCE, decreasing = TRUE),]
        
        # return everything
        return(list( fitted_test=fitted_test, importance=imp))
    }
    
    
    # The sprout_tree function samples the data set aside for training with replacement
    sprout_tree <- function(formula, m_try, data_train, node_size, data_test, approach) {
        
        #formula_train = formula for model fit
        #m_try = number of sampled features at each node for splitting
        #data_train = training data
        #data_test = test data
        #node_size = minimum node size
        #approach = "leaf" or "parent" depending on whether leaf or parent node is used to limit terminating node size (=node_size)
        
        
        # Set formula
        formula= formula_train
        
        # Extract all features
        features <- all.vars(formula)[-1]
        
        # Extract outcome
        outcome <- all.vars(formula)[1]
        
        # Randomly sample the data with replacement
        rand= sample(1:nrow(data_train), size = nrow(data_train), replace = TRUE)
        rand= sort(rand)
        
        # Data used to build model
        train <- data_train[c(rand),]
        n_train = nrow(data_train)
        rownames(train) = c(1:n_train)
        
        # Generate regression tree model and predictions
        treef=reg_tree(formula_train=formula_train , data=train , node_size=node_size , features=features, m_try=m_try, outcome=outcome, data_test=data_test, approach=approach)
        
        # Return predicted testing outcomes and the variable importances
        return(list( fits_test=treef$fitted_test, tree_imp=treef$importance  ))
    }
    
    
    # The reg_rf generates all trees in the RF model
    reg_rf <- function(formula, M, m_try, data_train, node_size, data_test, approach) {
        
        #formula_train = formula for model fit
        #M = number of trees
        #m_try = number of sampled features at each node for splitting
        #data_train = training data
        #data_test = test data
        #node_size = minimum node size
        #approach = "leaf" or "parent" depending on whether leaf or parent node is used to limit terminating node size
        
        
        
        # Generate all trees in the RF model
        trees <- plyr::raply(
        M,
        sprout_tree(
        formula = formula_train,
        m_try = m_try,
        data_train=data_train, node_size=node_size, data_test=data_test, approach=approach))
        
        
        # Collect all tree predictions
        fits_test <- do.call("cbind", trees[, 1])
        # Collect all tree variable importances
        imp_full <- do.call("rbind", trees[, 2])
        
        
        # Create sequence 1...., M (M= number of tree in RF)
        seq_trees= matrix(c(1:M), 1, M)
        
        # Take cumulative mean for each observation for m=1, ..., M
        predicted_cum= rowCumSum(fits_test)/col(rowCumSum(fits_test))
        
        # Final reported predicted value is for number of trees equal to M
        predicted_fins= as.matrix(predicted_cum[,M])
        
        # Return predicted testing outcomes and variable importances from M trees in the RF
        return(list(imp_full=imp_full,pred=predicted_fins))
    }
    
    
    # Set seed for reproducibility
    set.seed(seed_num)
    
    #Formula for fitting model
    formula_train = as.formula(paste0(colnames(training_data)[2], " ~ 0 +  ", paste0(colnames(training_data)[-c(1:2)], collapse =  " + ")))
    
    # Determine sample size for testing set
    n_test = dim(testing_data)[1]
    
    # Rename testing data
    data_test = testing_data
    
    # Assign rownames of testing set as 1, ..., n_test
    rownames(data_test) = c(1:n_test)
    
    # Determine sample size for testing set
    n_train= dim(training_data)[1]
    
    # Rename training data
    data_train = training_data
    
    # Assign rownames of training set as 1, ..., n_test
    rownames(data_train) = c(1:n_train)
    
    # Generate RF model
    RandForest_k= reg_rf(formula=formula_train, M=M, m_try=m_try, data_train=data_train, node_size=node_size, data_test=data_test, approach=approach)
    
    # Store variable importances from forest
    imp_full= RandForest_k$imp_full
    
    # Aggerate  variable importances from M trees in forest  by mean
    imp <- aggregate(IMPORTANCE ~ FEATURES, FUN = mean, imp_full)
    
    # Extract names of input variables
    names= colnames(data_train)[-c(1:2)]
    
    # Order importances by input variable name
    imp <- imp[order(factor(imp$FEATURES,levels=c(names))), ]
    
    # Store variable importances as data.frame
    var_imps= data.frame(imp)
    
    # Order importances from smallest to largest in order of influence
    data_imps <-  var_imps[order(var_imps$IMPORTANCE, decreasing = F),]
    
    # Generate variable importance plot
    var_plot= ggplot(data_imps, aes(x = reorder(FEATURES, -IMPORTANCE, decreasing = T), y = IMPORTANCE)) +
    geom_bar(stat = "identity")+
    coord_flip()+ labs(x = "Variables", y="Variable importances", title = "Variable importance plot")
    
    # Store testing predictions from RF
    pred_test= RandForest_k$pred
    
    # Compute weighted generalization error
    mse_test_weight= sum(data_test[,1]*(pred_test-data_test[,2])^2)/sum(data_test[,1])
    
    
    
    return(list("Estimated generalization error" =mse_test_weight, "Estimated variable importances" =data_imps, "Predicted testing outcome" =pred_test,"Testing outcome" =matrix(c(data_test[,2])), "m_try" =m_try, "node size" =node_size, "Number of trees" =M, "Approach used" =approach, var_plot))
    
}



############################################################################################################################### Illustration of using RF function

# Input name of training and testing sets in quotations:
# Load training set
training_data = read.csv("nhanes_16_RF.csv")

# Load testing set
testing_data = read.csv("nhanes_18_RF.csv")

M=500
m_try=3
node_size=20
training_data=training_data
testing_data=testing_data
approach= "leaf"
seed_num = 5

# Apply leaf approach having M=500 regression trees, m_try=3, node size=20, and random seed set to 5
results_leaf = runners(M=M, m_try=m_try, node_size=node_size, training_data=training_data, testing_data=testing_data, approach= approach, seed_num = seed_num)

# Resulting output
> results_leaf
$`Estimated generalization error`
[1] 0.1809785

$`Estimated variable importances`
FEATURES IMPORTANCE
16       quest_smoke_rec   52377.42
11      quest_doc_kidney   56349.29
5                 gender   59307.61
13   quest_risk_diabetes   92734.71
17      quest_stroke_med  120309.90
14        quest_salt_med  256190.25
18 quest_weight_loss_med  257128.21
12    quest_night_kidney  263284.42
15          quest_sedact  284898.68
19                  race  375951.23
10             lab_urine  391421.51
2         body_diastolic  432575.27
6                    inc  567026.22
4            body_weight  777384.90
7               lab_chol  829518.19
3          body_systolic 1177206.88
9               lab_trig 1380795.25
1                    age 3226089.37
8            lab_glucose 8991687.27


$`Predicted testing outcome`
[1,] 5.262604
[2,] 5.218241
[3,] 5.265470
[4,] 5.494832
[5,] 5.298605
.
.
.
[4048,] 5.592635
[4049,] 5.719622
[4050,] 5.606775
[4051,] 5.665789


$`Testing outcome`
[1,]  5.1
[2,]  4.9
[3,]  5.6
[4,]  5.1
[5,]  5.3
.
.
.
[4048,]  5.3
[4049,]  6.2
[4050,]  5.7
[4051,]  5.3

$m_try
[1] 3

$`node size`
[1] 20

$`Number of trees`
[1] 500

$`Approach used`
[1] "leaf"

# Predicted outcome vs Observed outcome plot
fit_plot = plot( results_leaf[[3]],results_leaf[[4]],
xlab="Predicted outcome",ylab="Observed outcome", xlim=c(min(results_leaf[[3]]),max(results_leaf[[3]])), ylim= c(min(results_leaf[[4]]),max(results_leaf[[4]])), main= "Observed outcome as a function of predicted outcome" )
abline(a=0,b=1)

# Residuals vs predicted outcome plot
res= results_leaf[[4]]-results_leaf[[3]]
plot(  results_leaf[[3]],res,
xlab="Predicted outcome",ylab="Residuals", main= "Residual plot: residuals as function of predicted outcome" )
abline(a=0,b=0)


